# -*- coding: utf-8 -*-

import elasticsearch_dsl
import json

from pybana.translators.elastic.buckets import BucketTranslator
from pybana.translators.elastic.metrics import MetricTranslator
from .filter import FilterTranslator

__all__ = ("ElasticTranslator", "FilterTranslator")


class ElasticTranslator:
    def translate(self, visualization, scope):
        """
        Transform a kibana visualization object into an elasticsearch_dsl Search.

        :param elasticsearch_dsl.Document visualization: Visualization fetched from a kibana index.
        :param Scope scope: Scope to use for data fetching.
        """
        index_pattern = visualization.index()
        index = index_pattern["index-pattern"]["title"]
        ts = index_pattern["index-pattern"]["timeFieldName"]
        search = elasticsearch_dsl.Search(index=index).filter(
            "range",
            **{ts: {"gte": scope.beg.isoformat(), "lte": scope.end.isoformat()}}
        )
        state = json.loads(visualization.visualization["visState"])
        segment_aggs = [
            agg
            for agg in state["aggs"]
            if agg["schema"] in ("segment", "group", "split", "bucket")
        ]
        metric_aggs = [agg for agg in state["aggs"] if agg["schema"] in ("metric",)]
        proxy = search.aggs
        for agg in segment_aggs:
            proxy = BucketTranslator().translate(proxy, agg, state, scope)
        for agg in metric_aggs:
            MetricTranslator().translate(proxy, agg, state)
        search = search[:0]
        search = search.filter(visualization.filters())
        return search



# -*- coding: utf-8 -*-

import json

"""
Provide translators which translate metric aggregations defined using kibana syntax to elasticsearch syntax.

Not supported:
- Top Hit
- Sibling pipeline aggregations
- Parent pipeline aggregations
"""


class BaseMetric:
    def params(self, agg):
        return json.loads(agg["params"].get("json") or "{}")

    def translate(self, proxy, agg, state):
        proxy.metric(agg["id"], agg["type"], **self.params(agg))


class AvgMetric(BaseMetric):
    aggtype = "avg"

    def params(self, agg):
        return {"field": agg["params"]["field"], **super().params(agg)}


class CardinalityMetric(BaseMetric):
    aggtype = "cardinality"

    def params(self, agg):
        return {"field": agg["params"]["field"], **super().params(agg)}


class CountMetric(BaseMetric):
    aggtype = "count"

    def translate(self, proxy, agg, state):
        pass


class MaxMetric(BaseMetric):
    aggtype = "max"

    def params(self, agg):
        return {"field": agg["params"]["field"], **super().params(agg)}


class MedianMetric(BaseMetric):
    aggtype = "median"

    def params(self, agg):
        return {
            "field": agg["params"]["field"],
            "percents": [50],
            **super().params(agg),
        }

    def translate(self, proxy, agg, state):
        proxy.metric(agg["id"], "percentiles", **self.params(agg))


class MinMetric(BaseMetric):
    aggtype = "min"

    def params(self, agg):
        return {"field": agg["params"]["field"], **super().params(agg)}


class PercentilesMetric(BaseMetric):
    aggtype = "percentiles"

    def params(self, agg):
        return {
            "field": agg["params"]["field"],
            "percents": agg["params"]["percents"],
            **super().params(agg),
        }


class PercentileRanksMetric(BaseMetric):
    aggtype = "percentile_ranks"

    def params(self, agg):
        return {
            "field": agg["params"]["field"],
            "values": agg["params"]["values"],
            **super().params(agg),
        }


class StdDevMetric(BaseMetric):
    aggtype = "std_dev"

    def params(self, agg):
        return {"field": agg["params"]["field"], **super().params(agg)}

    def translate(self, proxy, agg, state):
        proxy.metric(agg["id"], "extended_stats", **self.params(agg))


class SumMetric(BaseMetric):
    aggtype = "sum"

    def params(self, agg):
        return {"field": agg["params"]["field"], **super().params(agg)}


class DatasweetMetric(BaseMetric):
    aggtype = "datasweet_formula"

    def translate(self, proxy, agg, state):
        pass


TRANSLATORS = {
    translator.aggtype: translator
    for translator in (
        AvgMetric,
        CardinalityMetric,
        CountMetric,
        DatasweetMetric,
        MaxMetric,
        MedianMetric,
        MinMetric,
        PercentileRanksMetric,
        PercentilesMetric,
        StdDevMetric,
        SumMetric,
    )
}


class MetricTranslator:
    def translate(self, proxy, agg, state):
        TRANSLATORS[agg["type"]]().translate(proxy, agg, state)



# -*- coding: utf-8 -*-

__all__ = ("Scope",)


class Scope:
    """
    Scope associated to the visualization.

    :param beg datetime: Begin date of the period on which data should be fetched.
    :param end datetime: End date of the period on which data should be fetched.
    :param tzinfo (str, pytz.Timezone): Timezone of the request.
    :param config pybana.Config: Config of the kibana instance.
    """

    def __init__(self, beg, end, tzinfo, config):
        self.beg = beg
        self.end = end
        self.tzinfo = tzinfo
        self.config = config



# -*- coding: utf-8 -*-

import json

from .metrics import MetricTranslator

"""
Provide translators which translate bucket aggregations defined using kibana syntax to elasticsearch syntax.

Not supported:
- IPv4 range. Never used.
- Significant terms. Rarely used (maybe never).
- Terms:
  - "Group other values in separate bucket" is not handled. If set, kibana turns a "terms" to a "filters" agg.
  - Order by custom metric
"""


def compute_auto_interval(interval, beg, end):
    """
    Compute the automatic interval
    """
    if interval == "auto":
        delta = end - beg
        if delta.days >= 2 * 365:  # 2years
            return "30d"
        elif delta.days >= 365:
            return "1w"
        elif delta.days >= 48:
            return "1d"
        elif delta.days >= 12:
            return "12h"
        elif delta.days >= 4:
            return "3h"
        elif delta.total_seconds() >= 36 * 3600:
            return "1h"
        elif delta.total_seconds() >= 16 * 3600:
            return "30m"
        elif delta.total_seconds() >= 8 * 3600:
            return "10m"
        elif delta.total_seconds() >= 2 * 3600:
            return "5m"
        elif delta.total_seconds() >= 36 * 60:
            return "1m"
        elif delta.total_seconds() >= 12 * 60:
            return "30s"
        elif delta.total_seconds() >= 6 * 60:
            return "10s"
        elif delta.total_seconds() >= 4 * 60:
            return "5s"
        else:
            return "1s"
    return f"1{interval}"


def format_from_interval(interval):
    if interval.endswith("y"):
        return "yyyy"
    if interval.endswith("q") or interval.endswith("M"):
        return "yyyy-MM"
    if interval.endswith("d"):
        return "yyyy-MM-dd"
    if interval.endswith("h"):
        return "yyyy-MM-dd'T'HH'h'"
    return "date_time"


class BaseBucket:
    def translate(self, agg, state, context):
        return json.loads(agg["params"].get("json") or "{}")


class DateHistogramBucket(BaseBucket):
    aggtype = "date_histogram"

    def translate(self, agg, state, context):
        interval = compute_auto_interval(
            agg["params"]["interval"], context.beg, context.end
        )

        return {
            "field": agg["params"]["field"],
            "interval": interval,
            "time_zone": str(context.tzinfo),
            "format": format_from_interval(interval),
            **super().translate(agg, state, context),
        }


class DateRangeBucket(BaseBucket):
    aggtype = "date_range"

    def translate(self, agg, state, context):
        return {
            "field": agg["params"]["field"],
            "ranges": agg["params"]["ranges"],
            **super().translate(agg, state, context),
        }


class FiltersBucket(BaseBucket):
    aggtype = "filters"

    def translate(self, agg, state, context):
        filters = {}
        for fltr in agg["params"]["filters"]:
            label = fltr.get("label") or fltr["input"]["query"] or "*"
            filters[label] = (
                {
                    "query_string": {
                        "query": fltr["input"]["query"],
                        "analyze_wildcard": True,
                        "default_field": "*",
                    }
                }
                if fltr["input"]["query"]
                else {"match_all": {}}
            )
        return {"filters": filters, **super().translate(agg, state, context)}


class HistogramBucket(BaseBucket):
    aggtype = "histogram"

    def translate(self, agg, state, context):
        return {
            "field": agg["params"]["field"],
            "interval": agg["params"]["interval"],
            **super().translate(agg, state, context),
        }


class RangeBucket(BaseBucket):
    aggtype = "range"

    def translate(self, agg, state, context):
        return {
            "field": agg["params"]["field"],
            "ranges": agg["params"]["ranges"],
            **super().translate(agg, state, context),
        }


class TermsBucket(BaseBucket):
    aggtype = "terms"

    def translate(self, agg, state, context):
        orderby = agg["params"]["orderBy"]
        aggs = {agg["id"]: agg for agg in state["aggs"]}
        if orderby in aggs and aggs[orderby]["type"] == "count":
            orderby = "_count"
        return {
            "field": agg["params"]["field"],
            "size": agg["params"]["size"],
            "order": {orderby: agg["params"]["order"]},
            **super().translate(agg, state, context),
        }


TRANSLATORS = {
    translator.aggtype: translator
    for translator in (
        DateHistogramBucket,
        DateRangeBucket,
        FiltersBucket,
        HistogramBucket,
        RangeBucket,
        TermsBucket,
    )
}


class BucketTranslator:
    def translate(self, proxy, agg, state, context):
        ret = proxy.bucket(
            agg["id"],
            agg["type"],
            **TRANSLATORS[agg["type"]]().translate(agg, state, context),
        )
        for metric_agg in state["aggs"]:
            if metric_agg["id"] == agg["params"].get("orderBy"):
                MetricTranslator().translate(ret, metric_agg, state)
        return ret



# -*- coding: utf-8 -*-

import json
from elasticsearch import Elasticsearch
from elasticsearch_dsl import Search


def dumpindex(elastic, index, fn):
    """
    Helper which dump a whole db and returns in a format handled by bulk api
    """
    search = Search(using=elastic, index=index).filter(
        "terms", type=["index-pattern", "visualization", "dashboard", "search"]
    )
    with open(fn, "w+") as fd:
        for doc in search.scan():
            fd.write(
                json.dumps(
                    {
                        "_index": doc.meta.index,
                        "_type": doc.meta.doc_type,
                        "_id": doc.meta.id,
                        "_source": doc.to_dict(),
                    }
                )
            )
            fd.write("\n")


if __name__ == "__main__":
    import os

    dumpindex(
        Elasticsearch(),
        ".kibana_pybana",
        os.path.join(os.path.dirname(__file__), "index.json"),
    )



# -*- coding: utf-8 -*-

import ast
import re

__all__ = ("DatasweetTransformer",)


def is_variable(name):
    return bool(re.match("^agg\\d+$", name))


def ds_avg(*args):
    return sum(args) / len(args)


def ds_count(*args):
    return len(args)


def ds_cusum(*args):
    ret = []
    s = 0
    for a in args:
        s += a
        ret.append(s)
    return ret


def ds_derivative(*args):
    prev = float("nan")
    ret = []
    for arg in args:
        ret.append(arg - prev)
        prev = arg
    return ret


def ds_min(*args):
    return min(args)


def ds_max(*args):
    return max(args)


def ds_next(*args):
    return ds_prev(*args[::-1])[::-1]


def ds_prev(*args):
    prev = float("nan")
    ret = []
    for arg in args:
        ret.append(prev)
        prev = arg
    return ret


def ds_sum(*args):
    return sum(args)


FUNCS = {
    "avg": ds_avg,
    "count": ds_count,
    "cusum": ds_cusum,
    "derivative": ds_derivative,
    # "filter": TODO,
    # "if": TODO,
    # "ifnan": TODO,
    "min": ds_min,
    "max": ds_max,
    # "mvavg": TODO,
    "next": ds_next,
    "prev": ds_prev,
    "sum": ds_sum,
}


class DatasweetTransformer(ast.NodeTransformer):
    """
    Compile a datasweet formula

    - Make sure a only known names are used (for variables & funcs)
    - Rename `aggX` to `bucket["X"]["value"]`
    """

    def visit_Name(self, node):
        self.generic_visit(node)
        if node.id not in FUNCS and not is_variable(node.id):
            raise ValueError(f"{node.id} is not authorized")
        return node


def datasweet_eval(expr, bucket):
    tree = ast.parse(expr, mode="eval")
    tree = DatasweetTransformer().visit(tree)
    scope = {}
    for key, value in bucket.items():
        # TODO. Ugly. fix this.
        # count agg are not supported here
        if key.isdigit():
            val = (
                value["std_deviation"]
                if "std_deviation" in value
                else value["values"]["50.0"]
                if "values" in value
                else value["value"]
            )
            scope[f"agg{key}"] = float("nan") if val is None else val
    try:
        return eval(compile(tree, "a", mode="eval"), FUNCS, scope)
    except ZeroDivisionError:
        return None
